/*
Read infix expressions, evaluate them, and compare to expected value.
The program expects to be given a filename.  The file should contain
one or more sets of <infix expression> = <expected value>.  
*/

#define MAX_EXPRESSION_LENGTH 100

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "reporter.h"
#include "stack.h"
#include "postfix.h"

int main(int argc,char *argv[]) {

  FILE *src;			/* input FILE */
  int count;			/* number of values provided */
  int errors;			/* number of errors encountered */
  int status;			/* return function status */
  double value;			/* a single input value */
  double result;		/* result of evaluation */
  char infix[MAX_EXPRESSION_LENGTH+1];	/* infix expression buffer */
  char fmt[100];		/* format string for decoding input */
  Stack S;			/* pointer to stack object */
  char *postfix;		/* pointer to postfix expression */

  /* make sure that there is a filename */

  if (argc != 2) {
    printf("usage: %s <filename>\n",argv[0]);
    return 1;
  }

  /* open the file */

  src = fopen(argv[1],"r");
  if (src == NULL) {
    printf("file open error on %s.\n",argv[1]);
    return 1;
  }

  /* create a new stack */

  S = CreateStack(100);
  FatalErrorMemory(S);

  /* 
     Scan each line.  Read the expression, the '=', the expected
     value, and the terminating \n.
  */

  count = errors = 0;
  sprintf(fmt,"%%%i[^=]%%*s%%lf%%*[ \t\n\r\v\f]",MAX_EXPRESSION_LENGTH);
  while ((status = fscanf(src,fmt,infix,&value)) == 2) {
    printf("\ninfix:   %6.2f = %s\n",value,infix);
    postfix = ConvertToPostfix(infix,S);
    result = EvaluatePostfix(postfix,S);
    printf("postfix: %6.2f = %s\n",result,postfix);
    count++;
    if (fabs(result-value) > 1e-9) {
      printf("**Error** result != expected value.\n");
      errors++;
    }
    free(postfix);
  }

  printf("\n-------------\n\n");
  printf("%i cases, %i errors.\n",count,errors);
  fclose(src);

  DestroyStack(S);

  return 0;
}
