#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "reporter.h"
#include "postfix.h"

int isRightParen(char *s);
int isLeftParen(char *s);
int isOperator(char *s);
int isLowerPriority(char *a,char *b);
int isOperand(char *s);

double EvaluatePostfix(char *exp,Stack S) {
  struct operand {
    double v;
  };

  int pos;			/* position in expression */
  int delta;			/* characters read this scan */
  int status;			/* function status return */
  double result;		/* the result of the evaluation */
  char str[100];		/* one input item */
  struct operand *ea;		/* pointer to operand object */
  struct operand *eb;		/* pointer to operand object */

  /* parse the postfix string and evaluate it */
  
  pos = 0;
  while ((status = sscanf(&exp[pos],"%99s%n",str,&delta)) > 0) {
    pos += delta;
    if (isOperator(str)) {
      eb = Pop(S);
      ea = Pop(S);
      if (ea == NULL || eb == NULL) {
	printf("EvaluatePostFix: missing operand for %s\n",str);
	return 0.0;
      }
      if (strcmp(str,"+")==0) ea->v += eb->v;
      else if (strcmp(str,"-")==0) ea->v -= eb->v;
      else if (strcmp(str,"*")==0) ea->v *= eb->v;
      else if (strcmp(str,"/")==0) ea->v /= eb->v;
      else {
	printf("EvaluatePostFix: unrecognized operator: %s",str);
	return 0.0;
      }
      Push(S,ea);
      free(eb);
    } else {
      ea = malloc(sizeof(struct operand));
      FatalErrorMemory(ea);
      ea->v = atof(str);
      Push(S,ea);
    }
  }
  ea = Pop(S);
  if (ea == NULL) {
    printf("EvaluatePostFix: missing result.");
    return 0.0;
  }
  result = ea->v;
  free(ea);
  return result;
}      

/* 
read infix notation from string exp,
write postfix notation to new string,
using Stack S.
*/

char *ConvertToPostfix(char *exp,Stack S) {

  char *postfix;		/* pointer to the new result string */
  char *element;		/* pointer to a string */
  char *tos;			/* pointer to a string */
  int status;			/* function status return */
  char str[100];		/* one input item */
  int pos;			/* position in expression */
  int delta;			/* characters read this scan */

  /* create the result string */

  postfix = malloc(strlen(exp)+1);
  FatalErrorMemory(postfix);
  postfix[0] = '\0';

  /* parse the infix string and convert to postfix */
  
  pos = 0;
  while ((status = sscanf(&exp[pos],"%99s%n",str,&delta)) > 0) {
    pos += delta;
    if (isOperand(str)) {
      if (postfix[0] != '\0') strcat(postfix," ");
      strcat(postfix,str);
    } else if (isRightParen(str)) {
      while ((element = Pop(S)) != NULL && !isLeftParen(element)) {
	if (postfix[0] != '\0') strcat(postfix," ");
	strcat(postfix,element);
	free(element);
      }
    } else if (isOperator(str) || isLeftParen(str)) {
      while ((tos=TopOfStack(S)) != NULL && 
	     !isLeftParen(tos) &&
	     !isLowerPriority(tos,str)) {
	tos = Pop(S);
	if (postfix[0] != '\0') strcat(postfix," ");
	strcat(postfix,tos);
	free(tos);
      }
      element = strdup(str);
      FatalErrorMemory(element);
      Push(S,element);
    } else {
      printf("unrecognized element: %s\n",str);
    }
  }

  /* pop everything still on the stack */

  while ((element = Pop(S)) != NULL) {
    if (postfix[0] != '\0') strcat(postfix," ");
    strcat(postfix,element);
    free(element);
  }
  return postfix;
}

int isRightParen(char *s) {
  return (*s==')');
}

int isLeftParen(char *s) {
  return (*s=='(');
}

int isOperator(char *s) {
  return (strlen(s)==1) && 
    ((*s=='+') || (*s=='-') || (*s=='*') || (*s=='/'));
}

int isLowerPriority(char *a,char *b) {
  int pa;
  int pb;

  pa = 0;
  if (*a == '+' || *a == '-') pa = 0;
  else if (*a == '*' || *a == '/') pa = 1;
  else if (*a == '(') pa = 2;

  pb = 0;
  if (*b == '+' || *b == '-') pb = 0;
  else if (*b == '*' || *b == '/') pb = 1;
  else if (*b == '(') pb = 2;

  return pa < pb;
}

int isOperand(char *s) {
  return !(isOperator(s) || isLeftParen(s) || isRightParen(s));
}
